//Using SDL and standard IO
#include "SDL.h"
#include <cstdio>
#include <string>

//Screen dimension constants
const int SCREEN_WIDTH = 640;
const int SCREEN_HEIGHT = 480;
const float SPEED = 0.16f;
const float SMOOTH = 0.5f;

class Vector2D {
public:

	Vector2D(float x, float y) : m_x(x), m_y(y) {}
	float getX() { return m_x; }
	float getY() { return m_y; }

	void setX(float x) { m_x = x; }
	void setY(float y) { m_y = y; }

	float length() { return sqrt(m_x * m_x + m_y * m_y); }

	Vector2D operator+(const Vector2D& v2) const {
		return Vector2D(m_x + v2.m_x, m_y + v2.m_y);
	}

	friend Vector2D& operator+=(Vector2D& v1, const Vector2D& v2) {
		v1.m_x += v2.m_x;
		v1.m_y += v2.m_y;
		return v1;
	}

	Vector2D operator*(float scalar) {
		return Vector2D(m_x * scalar, m_y * scalar);
	}

	Vector2D& operator*=(float scalar) {
		m_x *= scalar;
		m_y *= scalar;
		return *this;
	}

	Vector2D operator-(const Vector2D& v2) const {
		return Vector2D(m_x - v2.m_x, m_y - v2.m_y);
	}
	friend Vector2D& operator-=(Vector2D& v1, const Vector2D& v2) {
		v1.m_x -= v2.m_x;
		v1.m_y -= v2.m_y;
		return v1;
	}

	Vector2D operator/(float scalar) {
		return Vector2D(m_x / scalar, m_y / scalar);
	}
	Vector2D& operator/=(float scalar) {
		m_x /= scalar;
		m_y /= scalar;
		return *this;
	}

	void normalize() {
		float l = length();
		if (l > 0) {
			(*this) *= 1 / l;
		}
	}

private:

	float m_x;
	float m_y;
};

//Texture wrapper class
class KTexture
{
public:
	//Initializes variables
	KTexture();

	//Deallocates memory
	~KTexture();

	//Loads image at specified path
	bool loadFromFile(std::string path);

	//Deallocates texture
	void free();

	//Renders texture at given point
	void render(int x, int y, SDL_Rect* clip = NULL);

	//Gets image dimensions
	int getWidth();
	int getHeight();

private:
	//The actual hardware texture
	SDL_Texture* mTexture;

	//Image dimensions
	int mWidth;
	int mHeight;
};

SDL_Window* gWindow = NULL;

SDL_Renderer* gRenderer = NULL;

KTexture gTexture1;
KTexture gTexture2;

bool init();

bool loadMedia();

void close();
SDL_Texture* loadTexture(std::string path);

KTexture::KTexture()
{
	//Initialize
	mTexture = NULL;
	mWidth = 0;
	mHeight = 0;
}

KTexture::~KTexture()
{
	//Deallocate
	free();
}

bool KTexture::loadFromFile(std::string path)
{
	//Get rid of preexisting texture
	free();
	//The final texture
	SDL_Texture* newTexture = NULL;

	//Load image at specified path
	SDL_Surface* loadedSurface = SDL_LoadBMP(path.c_str());
	if (loadedSurface == NULL)
	{
		printf("Unable to load image %s! SDL_image Error: %s\n", path.c_str(), SDL_GetError());
	}
	else
	{
		//Color key image
		SDL_SetColorKey(loadedSurface, SDL_TRUE, SDL_MapRGB(loadedSurface->format, 0xFF, 0, 0));
		//Create texture from surface pixels
		newTexture = SDL_CreateTextureFromSurface(gRenderer, loadedSurface);
		if (newTexture == NULL)
		{
			printf("Unable to create texture from %s! SDL Error: %s\n", path.c_str(), SDL_GetError());
		}
		else
		{
			//Get image dimensions
			mWidth = loadedSurface->w;
			mHeight = loadedSurface->h;
		}

		//Get rid of old loaded surface
		SDL_FreeSurface(loadedSurface);
	}

	//Return success
	mTexture = newTexture;
	return mTexture != NULL;
}

void KTexture::free()
{
	//Free texture if it exists
	if (mTexture != NULL)
	{
		SDL_DestroyTexture(mTexture);
		mTexture = NULL;
		mWidth = 0;
		mHeight = 0;
	}
}

void KTexture::render(int x, int y, SDL_Rect* clip)
{
	//Set rendering space and render to screen
	SDL_Rect renderQuad = { x, y, mWidth, mHeight };

	//Set clip rendering dimensions
	if (clip != NULL)
	{
		renderQuad.w = clip->w;
		renderQuad.h = clip->h;
	}

	//Render to screen
	SDL_RenderCopy(gRenderer, mTexture, clip, &renderQuad);
}

int KTexture::getWidth()
{
	return mWidth;
}

int KTexture::getHeight()
{
	return mHeight;
}

bool init()
{
	//Initialization flag
	bool success = true;

	//Initialize SDL
	if (SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		printf("SDL could not initialize! SDL_Error: %s\n", SDL_GetError());
		success = false;
	}
	else
	{
		//Set texture filtering to linear
		if (!SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "1"))
		{
			printf("Warning: Linear texture filtering not enabled!");
		}
		//Create window
		gWindow = SDL_CreateWindow("2DGK - Zadanie 2", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_SHOWN);
		if (gWindow == NULL)
		{
			printf("Window could not be created! SDL_Error: %s\n", SDL_GetError());
			success = false;
		}
		else
		{
			//Create renderer for window
			gRenderer = SDL_CreateRenderer(gWindow, -1, SDL_RENDERER_ACCELERATED);
			if (gRenderer == NULL)
			{
				printf("Renderer could not be created! SDL Error: %s\n", SDL_GetError());
				success = false;
			}
			else
			{
				//Initialize renderer color
				SDL_SetRenderDrawColor(gRenderer, 0, 0xFF, 0, 0xFF);
			}
		}
	}

	return success;
}

bool loadMedia()
{
	//Loading success flag
	bool success = true;


	if (!gTexture1.loadFromFile("textures/01.bmp"))
	{
		printf("Failed to load 01 texture image!\n");
		success = false;
	}

	if (!gTexture2.loadFromFile("textures/02.bmp"))
	{
		printf("Failed to load 01 texture image!\n");
		success = false;
	}

	return success;
}

SDL_Texture* loadTexture(std::string path)
{
	//The final optimized image
	SDL_Texture* newTexture = NULL;

	//Load image at specified path
	SDL_Surface* loadedSurface = SDL_LoadBMP(path.c_str());
	if (loadedSurface == NULL)
	{
		printf("Unable to load image %s! SDL Error: %s\n", path.c_str(), SDL_GetError());
	}
	else
	{
		//Create texture from surface pixels
		newTexture = SDL_CreateTextureFromSurface(gRenderer, loadedSurface);
		if (newTexture == NULL)
		{
			printf("Unable to optimize image %s! SDL Error: %s\n", path.c_str(), SDL_GetError());
		}

		//Get rid of old loaded surface
		SDL_FreeSurface(loadedSurface);
	}

	return newTexture;
}

void close()
{
	//Free loaded images
	gTexture1.free();
	gTexture2.free();

	//Destroy window    
	SDL_DestroyRenderer(gRenderer);
	SDL_DestroyWindow(gWindow);
	gWindow = NULL;
	gRenderer = NULL;

	//Quit SDL subsystems
	SDL_Quit();
}

int main(int argc, char* args[])
{
	//Start up SDL and create window
	if (!init())
	{
		printf("Failed to initialize!\n");
	}
	else
	{
		//Load media
		if (!loadMedia())
		{
			printf("Failed to load media!\n");
		}
		else
		{
			//Main loop flag
			bool quit = false;

			//Event handler
			SDL_Event e;

			Vector2D playerOnePosition(0, 0);
			Vector2D playerOneInput(0, 0);
			Vector2D playerOneVelocity(0, 0);
			Vector2D playerTwoPosition(0, 0);
			Uint8 previousKeyStates[] = {0,0,0,0};
			//While application is running
			while (!quit)
			{
				playerOneInput.setX(0);
				playerOneInput.setY(0);
				//Handle events on queue
				while (SDL_PollEvent(&e) != 0)
				{
					//User requests quit
					if (e.type == SDL_QUIT)
					{
						quit = true;
					}
					else if (e.type == SDL_MOUSEMOTION)
					{
						int x, y;
						SDL_GetMouseState(&x, &y);
						printf("[MOUSEMOTION]: x=%d, y=%d\n", x, y);
						playerTwoPosition.setX(x - gTexture2.getWidth() / 2);
						playerTwoPosition.setY(y - gTexture2.getHeight() / 2);
					}
				}

				const Uint8* currentKeyStates = SDL_GetKeyboardState(NULL);
				if (currentKeyStates[SDL_SCANCODE_W])
				{
					playerOneInput.setY(-1);
					printf("[KBD] W is pressed!\n");
				}
				if (currentKeyStates[SDL_SCANCODE_S])
				{
					playerOneInput.setY(1);
					printf("[KBD] S is pressed!\n");
				}
				if (currentKeyStates[SDL_SCANCODE_A])
				{
					playerOneInput.setX(-1);
					printf("[KBD] A is pressed!\n");
				}
				if (currentKeyStates[SDL_SCANCODE_D])
				{
					playerOneInput.setX(1);
					printf("[KBD] D is pressed!\n");
				}

				playerOneInput.normalize();
				playerOneVelocity = playerOneInput * SPEED * (1 - SMOOTH) + playerOneVelocity * SMOOTH;

				playerOnePosition += playerOneVelocity;

				if (playerOnePosition.getX() < 0) playerOnePosition.setX(0);
				if (playerOnePosition.getX() > SCREEN_WIDTH - gTexture1.getWidth()) playerOnePosition.setX(SCREEN_WIDTH - gTexture1.getWidth());
				if (playerOnePosition.getY() < 0) playerOnePosition.setY(0);
				if (playerOnePosition.getY() > SCREEN_HEIGHT - gTexture1.getHeight()) playerOnePosition.setY(SCREEN_HEIGHT - gTexture1.getHeight());


				//Clear screen
				SDL_SetRenderDrawColor(gRenderer, 0, 0xFF, 0, 0xFF);
				SDL_RenderClear(gRenderer);

				
				gTexture2.render(playerTwoPosition.getX(), playerTwoPosition.getY()); gTexture1.render(playerOnePosition.getX(), playerOnePosition.getY());

				



				//Update screen
				SDL_RenderPresent(gRenderer);
			}
		}
	}

	//Free resources and close SDL
	close();

	return 0;
}